@@ -13,7 +13,7 @@
 # under the License.
 
 """
-Checks code in this repository using various methods (MyPy, importing modules, syntax checks,
+Checks code in a repository using various methods (MyPy, importing modules, syntax checks,
 pycodestyle). Runs multiple checks in parallel.
 """
 
@@ -29,9 +29,6 @@ import logging
 import multiprocessing
 import re
 
-import configparser
-from configparser import ConfigParser
-
 from typing import List, Dict, Tuple, Set, Optional, Any
 
 from codecheck.check_result import CheckResult
@@ -39,95 +36,40 @@ from codecheck.reporter import Reporter
 from codecheck.util import (
     increment_counter,
     ensure_str_decoded,
-    combine_value_lists,
     get_module_name_from_path,
+    CompiledRE,
+    prepend_path_entries,
+)
+from codecheck.config import CodeCheckConfig
+from codecheck.constants import (
+    DEFAULT_CONF_FILE_NAME,
+    ALL_CHECK_TYPES,
+    ALL_CHECKED_SUFFIXES,
+    NAME_SUFFIX_TO_CHECK_TYPES,
 )
-
-NAME_SUFFIX_TO_CHECK_TYPES: Dict[str, List[str]] = {
-    '.py': ['mypy', 'compile', 'pycodestyle', 'doctest', 'import'],
-    '.sh': ['shellcheck'],
-    '_test.py': ['unittest'],
-}
-
-ALL_CHECK_TYPES: List[str] = combine_value_lists(NAME_SUFFIX_TO_CHECK_TYPES)
-
-DEFAULT_CONF_FILE_NAME = 'codecheck.ini'
 
 NUM_REMAINING_CHECKS_TO_SHOW = 5
 
+INDENTATION_SEPARATOR = '\n' + ' ' * 4
 
-def print_stats(description: str, d: Dict[str, int]) -> None:
-    print("%s:\n    %s" % (
+
+def print_stats(
+        description: str,
+        d: Dict[str, int],
+        failure_counts: Dict[str, int] = {}) -> None:
+    print("%s:%s%s" % (
         description,
-        '\n    '.join('%s: %s' % (k, v) for k, v in sorted(d.items()))
+        INDENTATION_SEPARATOR,
+        INDENTATION_SEPARATOR.join(
+            '%s: %s%s' % (
+                k,
+                v,
+                ' (%d failed)' % failure_counts[k] if k in failure_counts else ''
+            ) for k, v in sorted(d.items())
+        )
     ))
 
 
-if sys.version_info <= (3, 7):
-    CompiledRE = Any
-else:
-    CompiledRE = re.Pattern
-
-
-class CodeCheckConfig:
-    mypy_config_path: str
-    disabled_check_types: Set[str]
-    included_regex_list: Optional[List[CompiledRE]]
-
-    def __init__(self) -> None:
-        self.mypy_config_path = 'mypy.ini'
-        self.disabled_check_types = set()
-        self.included_regex_list = None
-
-    def load(self, file_path: str) -> None:
-        parsed_ini = ConfigParser()
-        parsed_ini.read(file_path)
-
-        def get_section(section_name: str) -> Optional[configparser.SectionProxy]:
-            if section_name in parsed_ini.sections():
-                return parsed_ini[section_name]
-            return None
-
-        def get_multi_line_regex_list(
-                section: configparser.SectionProxy,
-                field_name: str) -> List[CompiledRE]:
-            field_value = section.get(field_name)
-            if field_value is None:
-                return None
-            re_strings = field_value.strip().split('\n')
-            result: List[CompiledRE] = []
-            for re_str in re_strings:
-                re_str = re_str.strip()
-                if not re_str:
-                    continue
-                try:
-                    compiled_re = re.compile(re_str)
-                except Exception as ex:
-                    logging.exception("Failed to compile regular expression: %s", re_str)
-                    raise ex
-                result.append(compiled_re)
-
-            return result
-
-        default_section = get_section('default')
-        if default_section:
-            mypy_config_path = default_section.get('mypy_config')
-            if mypy_config_path is not None:
-                self.mypy_config_path = mypy_config_path
-
-        checks_section = get_section('checks')
-        if checks_section:
-            for check_type in ALL_CHECK_TYPES:
-                is_check_enabled = checks_section.get(check_type)
-                if is_check_enabled is not None and not checks_section.getboolean(check_type):
-                    self.disabled_check_types.add(check_type)
-
-        files_section = get_section('files')
-        if files_section:
-            self.included_regex_list = get_multi_line_regex_list(
-                files_section, 'included_regex_list')
-
-
 class CodeChecker:
     config: CodeCheckConfig
     args: argparse.Namespace
@@ -196,7 +138,6 @@ class CodeChecker:
         assert check_type in ALL_CHECK_TYPES
 
         append_file_path = True
-        rel_path = self.relativize_path(file_path)
 
         extra_messages = []
 
@@ -230,7 +171,6 @@ class CodeChecker:
             args = [self.args.python_interpreter, '-m', 'pycodestyle']
         elif check_type == 'unittest':
             append_file_path = False
-            rel_path_components = os.path.splitext(rel_path)[0].split('/')
             if fully_qualified_module_name is None:
                 raise ValueError(
                     'Could not identify the fully-qualified module name to use to invoke the '
@@ -246,6 +186,11 @@ class CodeChecker:
             args.append(file_path)
 
         subprocess_env = os.environ.copy()
+        if additional_sys_path:
+            for env_var_name in ['PYTHONPATH', 'MYPYPATH']:
+                subprocess_env[env_var_name] = prepend_path_entries(
+                    additional_sys_path, os.getenv(env_var_name)
+                )
 
         process = subprocess.Popen(
             args,
@@ -279,21 +224,39 @@ class CodeChecker:
             if self.args.verbose:
                 logging.info(f"Configuration file not found: {self.args.config_path}")
 
-    def filter_with_glob_patterns(
+    def filter_with_inclusion_exclusion_patterns(
             self, initial_list: List[str],
-            re_pattern_list: List[CompiledRE]) -> List[str]:
-        filtered_list = [
-            item for item in initial_list
-            if any(re_pattern.match(item) for re_pattern in re_pattern_list)
-        ]
+            re_pattern_list: List[Tuple[bool, CompiledRE]]) -> List[str]:
+        filtered_list = []
+        for item in initial_list:
+            should_include_item = False
+            # Try patterns one by one, and each pattern overrides the result of previous ones if
+            # it matches. So, we can include some files, then exclude some of those files, then
+            # again include some of the excluded files.
+            for is_inclusion_pattern, re_pattern in re_pattern_list:
+                if re_pattern.match(item):
+                    should_include_item = is_inclusion_pattern
+            if should_include_item:
+                filtered_list.append(item)
+
         if self.args.verbose:
             logging.info(
                 "Filtered %d files to %d using the regular expressions %s",
                 len(initial_list),
                 len(filtered_list),
-                [re_pattern.pattern for re_pattern in re_pattern_list])
+                [
+                    ('included' if is_inclusion_pattern else 'excluded', re_pattern.pattern)
+                    for is_inclusion_pattern, re_pattern in re_pattern_list
+                ])
         return filtered_list
 
+    def get_rel_dir_name_for_report(self, file_path: str) -> str:
+        """
+        Returns the directory name (potentially with multiple path components) of the given file
+        relative to the repository root. Used in the "checks by directory" section of the report.
+        """
+        return os.path.dirname(self.relativize_path(file_path)) or 'root'
+
     def run(self) -> bool:
         self.parse_args()
         self.init_config()
@@ -306,13 +269,12 @@ class CodeChecker:
         )).split('\n')
 
         if self.config.included_regex_list is not None:
-            file_list = self.filter_with_glob_patterns(
+            file_list = self.filter_with_inclusion_exclusion_patterns(
                 file_list, self.config.included_regex_list)
 
-        all_checked_suffixes = tuple(sorted(NAME_SUFFIX_TO_CHECK_TYPES.keys()))
         input_file_paths = set([
             os.path.abspath(file_path) for file_path in file_list if (
-                file_path.endswith(all_checked_suffixes)
+                file_path.endswith(ALL_CHECKED_SUFFIXES)
             )
         ])
 
@@ -336,11 +298,16 @@ class CodeChecker:
             )
 
         reporter = Reporter(line_width=80)
+
         checks_by_dir: Dict[str, int] = {}
+        checks_by_dir_failed: Dict[str, int] = {}
+
         checks_by_type: Dict[str, int] = {}
+        checks_by_type_failed: Dict[str, int] = {}
+
         checks_by_result: Dict[str, int] = {}
 
-        is_success = True
+        overall_success = True
 
         check_inputs = []
         if self.args.verbose:
@@ -348,7 +315,7 @@ class CodeChecker:
                 logging.info(f"Disabled check types: {sorted(self.config.disabled_check_types)}")
 
         for file_path in input_file_paths:
-            rel_dir = os.path.dirname(self.relativize_path(file_path)) or 'root'
+            rel_dir = self.get_rel_dir_name_for_report(file_path)
             for file_name_suffix, check_types in NAME_SUFFIX_TO_CHECK_TYPES.items():
                 for check_type in check_types:
                     if check_type in self.config.disabled_check_types:
@@ -373,21 +340,29 @@ class CodeChecker:
             num_completed = 0
             for future in concurrent.futures.as_completed(future_to_check_input):
                 file_path, check_type = future_to_check_input[future]
+
+                this_check_succeeded = True
                 try:
                     check_result = future.result()
                 except Exception as exc:
                     print(
                         f"Check '{check_type}' for '{file_path}' generated an exception: "
                         f"{traceback.format_exc()}")
-                    increment_counter(checks_by_result, 'failure')
-                    is_success = False
+                    this_check_succeeded = False
                 else:
                     reporter.print_check_result(check_result)
-                    if check_result.returncode == 0:
-                        increment_counter(checks_by_result, 'success')
-                    else:
-                        increment_counter(checks_by_result, 'failure')
-                        is_success = False
+                    if check_result.returncode != 0:
+                        this_check_succeeded = False
+
+                if this_check_succeeded:
+                    increment_counter(checks_by_result, 'success')
+                else:
+                    increment_counter(checks_by_result, 'failure')
+                    increment_counter(checks_by_type_failed, check_type)
+                    rel_dir = self.get_rel_dir_name_for_report(file_path)
+                    increment_counter(checks_by_dir_failed, check_type)
+                    overall_success = False
+
                 num_completed += 1
                 pending_checks.remove((file_path, check_type))
                 if self.args.detailed_progress:
@@ -403,22 +378,23 @@ class CodeChecker:
                         remaining_checks_to_show)
 
         if checks_by_dir:
-            print_stats("Checks by directory (relative to repo root)", checks_by_dir)
+            print_stats("Checks by directory (relative to repo root)",
+                        checks_by_dir, checks_by_dir_failed)
 
         if checks_by_type:
-            print_stats("Checks by type", checks_by_type)
+            print_stats("Checks by type", checks_by_type, checks_by_type_failed)
 
         if checks_by_result:
             print_stats("Checks by result", checks_by_result)
 
         print("Elapsed time: %.1f seconds" % (time.time() - start_time))
         print()
-        if is_success:
+        if overall_success:
             print(f"All {num_checks} checks are successful")
         else:
             print(f"Some checks failed")
         print()
-        return is_success
+        return overall_success
 
 
 def main() -> None:
