@@ -11,10 +11,17 @@
 # under the License.
 
 
-from typing import Dict, Union, Set, List
+from typing import Dict, Union, Set, List, Optional
 
 import os
 import sys
+import re
+
+
+if sys.version_info <= (3, 7):
+    CompiledRE = Any
+else:
+    CompiledRE = re.Pattern
 
 
 def increment_counter(d: Dict[str, int], key: str) -> None:
@@ -41,21 +48,22 @@ def get_module_name_from_path(file_path: str) -> str:
     return os.path.splitext(os.path.basename(file_path))[0]
 
 
-def is_valid_mypy_path_entry(path_entry: str) -> bool:
+def prepend_path_entries(new_entries: List[str], existing_path: Optional[str]) -> str:
     """
-    We try to configure MYPYPATH to be the same as PYTHONPATH, but without any site-packages
-    directories. Otherwise, mypy gives us the following message:
+    Prepend the given entries to the given PYTHONPATH or MYPYPATH-style string.
 
-    .../site-packages is in the MYPYPATH. Please remove it.
-    See https://mypy.readthedocs.io/en/latest/running_mypy.html#how-mypy-handles-imports for
-    more info.
-
-    Also we don't allow other system-wide module paths there.
+    >>> prepend_path_entries(['~/myproject'], None)
+    '~/myproject'
+    >>> prepend_path_entries(['~/myproject'], '')
+    '~/myproject'
+    >>> prepend_path_entries(['~/myproject'], '~/some_existing_dir')
+    '~/myproject:~/some_existing_dir'
     """
-    return (
-        os.path.basename(path_entry) != 'site-packages' and
-        '/site-packages/' not in path_entry and
-        # macOS-specific
-        '/Library/Frameworks/Python.framework/' not in path_entry and
-        not os.path.exists(os.path.join(path_entry, 'typing.py'))
-    )
+    if existing_path is None:
+        existing_path = ''
+    else:
+        existing_path = existing_path.strip()
+    if existing_path:
+        existing_path = ':' + existing_path
+    assert isinstance(new_entries, list), "Invalid list of new entries: %s" % new_entries
+    return ':'.join(new_entries) + existing_path
